API_CODES = {
    "Type": {
        "P": {
                "pl": "Typ jednostki – jednostka prawna",
                "eng": "Type of entity - juridical person"
            },
        "F": {
                "pl": "Typ jednostki – jednostka fizyczna (os. fizyczna prowadząca działalność gospodarczą)",
                "eng": "Type of entity - natural person"
            },
        "LP": {
                "pl": "Typ jednostki – jednostka lokalna jednostki prawnej",
                "eng": "Type of entity - local entity of juridical person"
            },
        "LF": {
                "pl": "Typ jednostki – jednostka lokalna jednostki fizycznej",
                "eng": "Type of entity - local entity of natural person"
            }
    },
    "SilosID": {
        "1": {
                "pl": "Miejsce prowadzenia działalności zarejestrowanej w CEIDG (tylko typy F i LF))",
                "eng": "Registered in CEIDG (only types F and LF)"
            },
        "2": {
                "pl": "Miejsce prowadzenia działalności rolniczej (tylko typy F i LF))",
                "eng": "Agricultural (only types F and LF)"
            },
        "3": {
                "pl": "Miejsce prowadzenia działalności pozostałej (tylko typy F i LF))",
                "eng": "Other activity (only types F and LF)"
            },
        "4": {
                "pl": "Miejsce prowadzenia działalności zlikwidowanej w starym systemie KRUPGN",
                "eng": "Liquidated (old KRUPGN system)"
            },
        "6": {
                "pl": "Miejsce prowadzenia działalności jednostki prawnej (tylko typy P i LP)",
                "eng": "Juridical person (only types P and LP)"
            }
    },
    "GetValue": {
        "0": {
                "pl": "Brak wiadomości",
                "eng": "No messages"
            },
        "1": {
                "pl": "Konieczne jest pobranie i sprawdzenie kodu Captcha (metody PobierzCaptcha i SprawdzCaptcha).",
                "eng": "It's necessary to download and check Captcha code (methods PobierzCaptcha i SprawdzCaptcha)"
            },
        "2": {
                "pl": "Do metody DaneSzukaj przekazano zbyt wiele identyfikatorów.",
                "eng": "Too much parameters passed to DaneSzukaj method"
            },
        "4": {
                "pl": "Nie znaleziono podmiotów.",
                "eng": "Can't find entity"
            },
        "5": {
                "pl": "Brak uprawnień do raportu.",
                "eng": "Insufficient privileges for this report"
            },
        "7": {
                "pl": "Brak sesji. Sesja wygasła lub przekazano nieprawidłową wartość nagłówka sid.",
                "eng": "Session failure. Session expired or sid header is incorrect"
            }
    },
    "ServiceStatus": {
        "0": {
                "pl": "Niedostępna",
                "eng": "Unavailable"
        },
        "1": {
                "pl": "Dostępna",
                "eng": "Available"
        },
        "2": {
                "pl": "Przerwa techniczna",
                "eng": "Maintenance"
        }
    }
}



"""
    Extends RegonAPI with operations
"""

from .parsers import parse_xml_response
from .exceptions import ApiUnknownReportNameError
from . import validators
from .settings import OPERATIONS


class RegonAPIOperations(object):
    def searchData(self, krs=None, regon=None, nip=None, regons9=None,
                   regons14=None, krss=None, nips=None):
        """Search data

        Parameters
        ----------
        krs : str
            KRS number
        regon : str
            REGON9 or REGON14
        nip : str
            NIP number
        regons9 : list
            List of REGON9 numbers
        regons14 : list
            List of REGON14 numbers
        krss : list
            List of KRS numbers
        nips : list
            List of NIP numbers

        Returns
        -------
        None
            If not found any results for provided parameters
        list
            List of results

        Raises
        ------
        Exception
            If no parameters provided
        TypeError
            If at least one parameter is provided with incorrect type
        """
        # Validate parameters
        str_params = [krs, regon, nip]
        list_params = [regons9, regons14, krss, nips]
        if not any(str_params) and not any(list_params):
            raise Exception("provide at least one parameter")

        if krs is not None and not isinstance(krs, str):
            raise TypeError("krs - invalid (str required)")
        if regon is not None and not isinstance(regon, str):
            raise TypeError("regon - invalid (str required)")
        if nip is not None and not isinstance(nip, str):
            raise TypeError("nip - invalid (str required)")

        if regons9 is not None and not isinstance(regons9, list):
            raise TypeError("regons9 - invalid (list required)")
        if regons14 is not None and not isinstance(regons14, list):
            raise TypeError("regons14 - invalid (list required)")
        if krss is not None and not isinstance(krss, list):
            raise TypeError("krss - invalid (list required)")
        if nips is not None and not isinstance(nips, list):
            raise TypeError("nips - invalid (list required)")

        if regons9 is not None:
            for _ in regons9:
                validators.is_valid_regon9(_)

        if regons14 is not None and not isinstance(regons14, list):
            for _ in regons9:
                validators.is_valid_regon14(_)

        # NIP validation

        # KRS validation

        # Validate parameters - End

        # join lists
        regons9 = ",".join(regons9) if regons9 else None
        regons14 = ",".join(regons14) if regons14 else None
        krss = ",".join(krss) if krss else None
        nips = ",".join(nips) if nips else None
        # join lists - End

        search_param = "pParametryWyszukiwania"
        request_data = {search_param: {}}
        map_ = {
            "krs": "Krs",
            "regon": "Regon",
            "nip": "Nip",
            "regons9": "Regony9zn",
            "regons14": "Regony14zn",
            "krss": "Krsy",
            "nips": "Nipy"
        }
        for k, v in locals().items():
            if k in map_.keys():
                if v is not None:
                    request_data[search_param][map_[k]] = v

        wsdl_method = getattr(
            self.service,
            OPERATIONS["alias_search_data"][self.bir_version]
        )
        response = wsdl_method(**request_data)
        return parse_xml_response(response) if response else None

    def dataDownloadFullReport(self, regon, report_name, strict=True):
        """Search data (DaneSzukaj wrapper)

        Parameters
        ----------
        regon : str
            REGON9 or REGON14
        report_name : str
            report name
        strict : bool
            If True checks if report_name is valid (default: True)

        Returns
        -------
        None
            If not found any results for provided parameters
        list
            List of results

        Raises
        ------
        ApiUnknownReportNameError
            If strict=True and provided report name is not included in
            predefined list of valid report names
        """
        if strict is True and report_name not in self.reports:
            raise ApiUnknownReportNameError(report_name)
        request_data = {
            "pRegon": regon,
            "pNazwaRaportu": report_name
        }
        wsdl_method = getattr(
            self.service,
            OPERATIONS["alias_data_download_full_report"][self.bir_version]
        )
        response = wsdl_method(**request_data)
        return parse_xml_response(response) if response else None



import pytest

from RegonAPI.exceptions import RegonConvertionError
from RegonAPI.converters import regon13_to_14
from RegonAPI.converters import regon8_to_9

testing = {
    "REGON13": '1234567851234',
    "REGON14": '12345678512347',
    "REGON8": '49270733',
    "REGON9": '492707333'
}


@pytest.mark.first
def test_converters_Prerequisites():
    assert testing["REGON8"] is not None
    assert testing["REGON9"] is not None
    assert testing["REGON13"] is not None
    assert testing["REGON14"] is not None


def test_CorrectRegon8Param_CorrectConversion():
    assert regon13_to_14(regon13=testing["REGON13"]) == testing["REGON14"]


def test_IncorrectRegon13Param_ExceptionRaised():
    try:
        regon13_to_14(regon13="testing")
        assert False
    except RegonConvertionError:
        assert True
    except Exception:
        assert False


def test_IncorrectParamType_ExceptionRaised():
    try:
        regon13_to_14(regon13=0)
        assert False
    except RegonConvertionError:
        assert True
    except Exception:
        assert False


def test_CorrectRegon8Param_CorrectConversion():
    """ Tests if convertion is working correctly """
    assert regon8_to_9(regon8=testing["REGON8"]) == testing["REGON9"]


def test_IncorrectRegon8Param_ExceptionRaised():
    try:
        regon8_to_9(regon8="testing")
        assert False
    except RegonConvertionError:
        assert True
    except Exception:
        assert False


def test_IncorrectParamType_ExceptionRaised():
    try:
        regon8_to_9(regon8=0)
        assert False
    except RegonConvertionError:
        assert True
    except Exception:
        assert False



from unittest import mock
import pytest

from RegonAPI.exceptions import ApiUnknownReportNameError
from RegonAPI import RegonAPI

testing = {
    "REGON9": "492707333",
    "REGON14": "12345678512347",
    "test_params": {
        "searchData": {
            "params": {
                "valid": [
                    {"krs": "0000006865"},
                    {"regon": "492707333"},
                    {"nip": "7342867148"},
                    {"regons9": ["492707333"]},
                    {"regons14": ["12345678512347"]},
                    {"krss": ["0000006865"]},
                    {"nips": ["7342867148"]},
                ],
                "invalid": [
                    {},
                    {
                        "krs": 123,
                        "regon": 123,
                        "nip": 123,
                        "regons9": "testing",
                        "regons14": "testing",
                        "krss": "testing",
                        "nips": "testing"
                    },
                ]
            }
        }
    }
}


@pytest.fixture
@mock.patch('RegonAPI.regon_api.Client', autospec=True)
def api_mock(mockClient):
    """ Creates api object with mocked client """
    def fin():
        pass
    api = RegonAPI()
    return api


@pytest.mark.first
def test_operations_Prerequisites():
    assert api_mock is not None
    assert testing is not None
    assert testing["test_params"] is not None
    assert testing["REGON9"] is not None
    assert testing["REGON14"] is not None


@mock.patch('RegonAPI.operations.parse_xml_response')
def test_dataDownloadFullReport_method__valid_calls(
    mocked_parse_xml_response,
    api_mock
):
    mocked_parse_xml_response.return_value = "testing"

    calls_made = len(api_mock.reports) * 2

    for _ in api_mock.reports:
        assert api_mock.dataDownloadFullReport(testing["REGON9"], _) == "testing"
        assert api_mock.dataDownloadFullReport(testing["REGON14"], _) == "testing"

    assert mocked_parse_xml_response.call_count == calls_made
    assert api_mock.service.DanePobierzPelnyRaport.call_count == calls_made


@mock.patch('RegonAPI.operations.parse_xml_response')
def test_dataDownloadFullReport_method__invalid_report_name(
    mocked_parse_xml_response,
    api_mock
):
    mocked_parse_xml_response.return_value = "testing"

    try:
        api_mock.dataDownloadFullReport(testing["REGON9"], "testing")
        assert False
    except ApiUnknownReportNameError as e:
        assert True


@mock.patch('RegonAPI.operations.parse_xml_response')
def test_searchData_method__valid_calls(mocked_parse_xml_response, api_mock):
    mocked_parse_xml_response.return_value = "testing"

    params = testing["test_params"]["searchData"]["params"]
    calls_made = len(params["valid"])

    for tc in params["valid"]:
        assert api_mock.searchData(**tc) == "testing"
    mocked_parse_xml_response.call_count == calls_made
    api_mock.service.DaneSzukaj.call_count == calls_made


@mock.patch('RegonAPI.operations.parse_xml_response')
def test_searchData_method__invalid_calls(mocked_parse_xml_response, api_mock):
    mocked_parse_xml_response.return_value = "testing"

    params = testing["test_params"]["searchData"]["params"]

    for tc in params["invalid"]:
        try:
            api_mock.searchData(**tc)
        except TypeError as e:
            assert True
        except Exception as e:
            assert True



import pytest

from RegonAPI.exceptions import ApiCodeTranslationError
from RegonAPI.settings import available_languages
from RegonAPI.settings import API_CODES
from RegonAPI.api_codes import t


@pytest.mark.first
def test_api_codes_Prerequisites():
    assert API_CODES is not None
    assert available_languages is not None


def test_t_CorrectOwnMapProvided_NoProblemsExpected():
    for method_key, method_value in API_CODES.items():
        testing_method = method_key
        for code in method_value.keys():
            try:
                # For settings.lang
                ret = t(method_key, code)
                assert isinstance(ret, str) or isinstance(ret, dict)
                # For forced lang
                for lang in available_languages:
                    ret = t(method_key, code, force_lang=lang)
                    assert isinstance(ret, str)
                    assert True
            except Exception as e:
                assert False


def test_t_CorrectParamsProvided_NoExceptionRaised():
    try:
        t(method="GetValue", code="4")
        assert True
    except Exception:
        assert False


def test_t_IncorrectCodeParamProvided_ExceptionRaised():
    try:
        t(method="GetValue", code="testing")
        assert False
    except ApiCodeTranslationError:
        assert True
    except Exception:
        assert False


def test_t_IncorrectMethodParamProvided_ExceptionRaised():
    try:
        t(method="testing", code="4")
        assert False
    except ApiCodeTranslationError:
        assert True
    except Exception:
        assert False


def test_t_IncorrectForceLangParamProvided_ExceptionRaised():
    try:
        t(method="GetValue", code="4", force_lang="testing")
        assert False
    except ApiCodeTranslationError:
        assert True
    except Exception:
        assert False


def test_t_CorrectForceLangParamProvided_NoExceptionRaised():
    try:
        t(method="GetValue", code="4", force_lang="pl")
        assert True
    except ApiCodeTranslationError:
        assert False



# Test API key
API_KEY_TEST_ENV = "abcde12345abcde12345"

# Warnings
WARNINGS = {
    "WARN_IS_NOT_PRODUCTION": {
        "eng": """
        \n\r\tYou are using TEST environment.
        \r\tSet `is_production=True` while constructing
        \r\t`RegonAPI` instance to switch to production environment.

        \r\tUsing test API key: '{API_KEY_TEST_ENV}'
        \r\tKey provided to `authenticate` method will be overridden.\n
        """.format(API_KEY_TEST_ENV=API_KEY_TEST_ENV),
        "pl": """
        \n\r\tUżywasz środowiska TESTOWEGO.
        \r\tUstaw `is_production=True` podczas konstrukcji
        \r\tinstancji klasy `RegonAPI` aby przejść do środowiska produkcyjnego.

        \r\tUżywany jest klucz testowy: '{API_KEY_TEST_ENV}'
        \r\tKlucz przekazany do metody `authenticate` zostanie nadpisany.\n
        """.format(API_KEY_TEST_ENV=API_KEY_TEST_ENV),
    }
}



from .voivodeships import VOIVODESHIPS
from .api_codes import API_CODES
from .bir_version_maps import *
from .dev_environment import API_KEY_TEST_ENV
from .dev_environment import WARNINGS as DEV_ENV_WARNINGS



import pytest

from RegonAPI.parsers import parse_xml_response


testing = {
    "parse_xml_response": {
        "params": {
            "valid": [
                """
                    <root>
                        <dane>
                            <test_param_1>val1</test_param_1>
                            <test_param_2>val2</test_param_2>
                            <test_param_3>val3</test_param_3>
                        </dane>
                        <dane>
                            <test_param_1>val4</test_param_1>
                            <test_param_2>val5</test_param_2>
                            <test_param_3>val6</test_param_3>
                        </dane>
                    </root>
                """
            ],
            "invalid": [
                "<root><dane><a>val1</a></dane></Root>",
                "<root><dane><a>val1</a></dane><root>",
                "<root><da><a>val1</a></dane></root>",
                "<dane><a>val1</a></dane>",
                "<root></root>",
                "<root><dane><b>val1</a></dane></root>",
                "<root><dane><a>val1</a><dane><root>",
                "<root><dane>a>val1</a></dane></root>",
                "",
                None,
                0,
                123,
                "testing123"
            ]
        }
    }
}


@pytest.mark.first
def test_parsers_Prerequisites():
    assert True is not None


def test_parse_xml_response__function_valid_call():
    params = testing["parse_xml_response"]["params"]
    result = parse_xml_response(params["valid"][0])
    assert result[0]["test_param_1"] == "val1"
    assert result[0]["test_param_2"] == "val2"
    assert result[0]["test_param_3"] == "val3"
    assert result[1]["test_param_1"] == "val4"
    assert result[1]["test_param_2"] == "val5"
    assert result[1]["test_param_3"] == "val6"


def test_parse_xml_response__function_invalid_calls():
    params = testing["parse_xml_response"]["params"]
    for _ in params["invalid"]:
        try:
            parse_xml_response(_)
            assert False
        except Exception as e:
            assert True


def test_parse_xml_response__function_invalid_call():
    try:
        parse_xml_response(None)
        assert False
    except TypeError:
        assert True



"""
    Translation of API codes to available languages
"""

import json

from . import settings
from .settings import API_CODES
from .exceptions import ApiCodeTranslationError


def t(method, code, force_lang=None):
    """Translates api code response to message

    Parameters
    ----------
    method : str
        Regon API operation (ex. GetValue)
    code : str
        API code response (ex. '1', '2', ...)
    force_lang: str
        forces function to translate to provided language despite settings

    Returns
    -------
    str
        if settings.lang is not "all".
        Translated message.
    dict
        if settings.lang is "all".
        Translated message in all available languages.

    Raises
    ------
    ApiCodeTranslationError
        if translation couldn't be performed
    """
    lang = settings.lang if force_lang is None else force_lang
    # Get by method
    msg = API_CODES.get(method, None)
    if msg is None:
        raise ApiCodeTranslationError("invalid method: %s" % method)
    # Get by code
    msg = msg.get(code, None)
    if msg is None:
        raise ApiCodeTranslationError("invalid code: %s" % code)

    if lang == "all":
        return msg

    msg = msg.get(lang, None)
    if msg is None:
        raise ApiCodeTranslationError("%s->%s for lang: '%s'" % (
            method, code, lang))

    return msg



BIR_SETTINGS = {
    "bir1": {  # BIR1
        "TEST": {
            "SERVICE_URL": "https://wyszukiwarkaregontest.stat.gov.pl/wsBIR/UslugaBIRzewnPubl.svc",
            "WSDL": "https://wyszukiwarkaregontest.stat.gov.pl/wsBIR/wsdl/UslugaBIRzewnPubl.xsd"
        },
        "PROD": {
            "SERVICE_URL": "https://wyszukiwarkaregon.stat.gov.pl/wsBIR/UslugaBIRzewnPubl.svc",
            # Yes, it's the same WSDL address for test and production in version 1.0
            "WSDL": "https://wyszukiwarkaregontest.stat.gov.pl/wsBIR/wsdl/UslugaBIRzewnPubl.xsd"
        }
    },
    "bir1.1": {  # BIR1.1
        "TEST": {
            "SERVICE_URL": "https://wyszukiwarkaregontest.stat.gov.pl/wsBIR/UslugaBIRzewnPubl.svc",
            "WSDL": "https://wyszukiwarkaregontest.stat.gov.pl/wsBIR/wsdl/UslugaBIRzewnPubl-ver11-test.wsdl"
        },
        "PROD": {
            "SERVICE_URL": "https://wyszukiwarkaregon.stat.gov.pl/wsBIR/UslugaBIRzewnPubl.svc",
            # Sneaky correction in version 1.1
            "WSDL": "https://wyszukiwarkaregon.stat.gov.pl/wsBIR/wsdl/UslugaBIRzewnPubl-ver11-prod.wsdl"
        }
    }
}

BIR_VERSIONS = BIR_SETTINGS.keys()

OPERATIONS = {
    "alias_search_data": {
        "bir1": "DaneSzukaj",
        "bir1.1": "DaneSzukajPodmioty"
    },
    "alias_data_download_full_report": {
        "bir1": "DanePobierzPelnyRaport",
        "bir1.1": "DanePobierzPelnyRaport"
    },
    "alias_data_download_full_group_report": {
        "bir1": NotImplementedError,
        "bir1.1": "DanePobierzRaportZbiorczy"
    }
}

# DanePobierzPelnyRaport - settings
DANE_POBIERZ_PELNY_RAPORT_REPORT_NAMES = [
    {
         "bir1": "PublDaneRaportFizycznaOsoba",
         "bir1.1": "BIR11OsFizycznaDaneOgolne"
    },
    {
         "bir1": "PublDaneRaportDzialalnoscFizycznejCeidg",
         "bir1.1": "BIR11OsFizycznaDzialalnoscCeidg"
    },
    {
         "bir1": "PublDaneRaportDzialalnoscFizycznejRolnicza",
         "bir1.1": "BIR11OsFizycznaDzialalnoscRolnicza"
    },
    {
         "bir1": "PublDaneRaportDzialalnoscFizycznejPozostala",
         "bir1.1": "BIR11OsFizycznaDzialalnoscPozostala"
    },
    {
         "bir1": "PublDaneRaportDzialalnoscFizycznejWKrupgn",
         "bir1.1": "BIR11OsFizycznaDzialalnoscSkreslonaDo20141108"
    },
    {
         "bir1": "PublDaneRaportDzialalnosciFizycznej",
         "bir1.1": "BIR11OsFizycznaPkd"
    },
    {
         "bir1": "PublDaneRaportLokalneFizycznej",
         "bir1.1": "BIR11OsFizycznaListaJednLokalnych"
    },
    {
         "bir1": "PublDaneRaportLokalnaFizycznej",
         "bir1.1": "BIR11JednLokalnaOsFizycznej"
    },
    {
         "bir1": "PublDaneRaportDzialalnosciLokalnejFizycznej",
         "bir1.1": "BIR11JednLokalnaOsFizycznejPkd"
    },
    {
         "bir1": "PublDaneRaportPrawna",
         "bir1.1": "BIR11OsPrawna"
    },
    {
         "bir1": "PublDaneRaportDzialalnosciPrawnej",
         "bir1.1": "BIR11OsPrawnaPkd"
    },
    {
         "bir1": "PublDaneRaportLokalnePrawnej",
         "bir1.1": "BIR11OsPrawnaListaJednLokalnych"
    },
    {
         "bir1": "PublDaneRaportLokalnaPrawnej",
         "bir1.1": "BIR11JednLokalnaOsPrawnej"
    },
    {
         "bir1": "PublDaneRaportDzialalnosciLokalnejPrawnej",
         "bir1.1": "BIR11JednLokalnaOsPrawnejPkd"
    },
    {
         "bir1": "PublDaneRaportWspolnicyPrawnej",
         "bir1.1": "BIR11OsPrawnaSpCywilnaWspolnicy"
    },
    {
         "bir1": "PublDaneRaportTypJednostki",
         "bir1.1": "BIR11TypPodmiotu"
    }
]

REPORTS = {
    "bir1": [x["bir1"] for x in DANE_POBIERZ_PELNY_RAPORT_REPORT_NAMES],
    "bir1.1": [x["bir1.1"] for x in DANE_POBIERZ_PELNY_RAPORT_REPORT_NAMES]
}



"""
    Parsing functions
"""

from bs4 import BeautifulSoup


def parse_xml_response(response):
    """Parses XML response to Python dict

    Parameters
    ----------
    response : str
        XML string response from Regon API operation

    Returns
    -------
    dict
        XML data converted to dict.

    Raises
    ------
    TypeError
        If response is not a str
    """
    if not isinstance(response, str):
        raise TypeError("parse_xml_response - response")
    soup = BeautifulSoup(response, 'xml')
    ret = []
    # Parse data
    filter_arr = ['\r\n', '\n', '\r']
    for dane in soup('dane'):
        children = list(filter(lambda x: x not in filter_arr, dane.children))
        dict_elem = {}
        for _ in children:
            dict_elem[_.name] = _.get_text()
        ret.append(dict_elem.copy())

    return ret



"""
    Converters of codes
"""

from string import digits

from .exceptions import RegonConvertionError
from . import validators


REGON9_WEIGHTS = [8, 9, 2, 3, 4, 5, 6, 7]
REGON14_WEIGHTS = [2, 4, 8, 5, 0, 9, 7, 3, 6, 1, 2, 4, 8]


def regon8_to_9(regon8):
    """Convert REGON8 to REGON9

    Parameters
    ----------
    regon8 : str
        REGON8

    Returns
    -------
    str
        REGON9

    Raises
    ------
    RegonConvertionError
        If regon8 is not valid
    """
    if not validators.is_valid_regon8(regon8):
        raise RegonConvertionError(regon8)
    a, b = list(regon8), REGON9_WEIGHTS
    a = list(map(lambda x: int(x), a))
    last_digit = sum(list(map(lambda x: x[0]*x[1], zip(a, b)))) % 11
    regon9 = "{regon8}{last_digit}".format(
        regon8=regon8,
        last_digit=last_digit)
    return regon9


def regon13_to_14(regon13):
    """Convert REGON13 to REGON14

    Parameters
    ----------
    regon13 : str
        REGON13

    Returns
    -------
    str
        REGON14

    Raises
    ------
    RegonConvertionError
        If regon13 is not valid
    """
    if not validators.is_valid_regon13(regon13):
        raise RegonConvertionError(regon13)
    a, b = list(regon13), REGON14_WEIGHTS
    a = list(map(lambda x: int(x), a))
    last_digit = sum(list(map(lambda x: x[0]*x[1], zip(a, b)))) % 11
    regon14 = "{regon13}{last_digit}".format(
        regon13=regon13,
        last_digit=last_digit)
    return regon14



"""
    Validator functions
"""

import re

from . import converters


def _re_is_digit_string(str_, str_len):
    """Digit string validation

    Parameters
    ----------
    str_ : str
        Digit string to validate
    str_len : int
        Digit string expected length

    Returns
    -------
    bool
        True if valid, False otherwise

    Raises
    ------
    TypeError
        If str_len is not positive int
    """
    regex = "^[0-9]{%s}$" % str_len
    if not isinstance(str_, str):
        return False
    if not isinstance(str_len, int):
        raise TypeError("_re_is_digit_string - str_len is not int")
    if str_len < 0:
        raise Exception("_re_is_digit_string - str_len < 0")
    ret = re.match(regex, str_)
    return True if ret is not None else False


def is_valid_regon8(regon8):
    """Regon8 validation

    Parameters
    ----------
    regon8 : str
        REGON8

    Returns
    -------
    bool
        True if valid, False otherwise

    Raises
    ------
    TypeError
        If regon8 is not str type
    """
    ret = _re_is_digit_string(regon8, 8)
    return True if ret else False


def is_valid_regon9(regon9):
    """Regon9 validation

    Parameters
    ----------
    regon9 : str
        REGON9

    Returns
    -------
    bool
        True if valid, False otherwise

    Raises
    ------
    TypeError
        If REGON9 is not str type
    """
    if not _re_is_digit_string(regon9, 9):
        return False
    regon8 = regon9[:8]
    return converters.regon8_to_9(regon8) == regon9


def is_valid_regon13(regon13):
    """REGON13 validation

    Parameters
    ----------
    regon13 : str
        REGON13

    Returns
    -------
    bool
        True if valid, False otherwise

    Raises
    ------
    TypeError
        If REGON13 is not str type
    """
    ret = _re_is_digit_string(regon13, 13)
    return True if ret else False


def is_valid_regon14(regon14):
    """REGON14 validation

    Parameters
    ----------
    regon14 : str
        REGON14

    Returns
    -------
    bool
        True if valid, False otherwise

    Raises
    ------
    TypeError
        If REGON14 is not str type
    """
    if not _re_is_digit_string(regon14, 14):
        return False
    regon13 = regon14[:13]
    return converters.regon13_to_14(regon13) == regon14



import pytest

from RegonAPI.validators import _re_is_digit_string
from RegonAPI.validators import is_valid_regon8
from RegonAPI.validators import is_valid_regon9
from RegonAPI.validators import is_valid_regon13
from RegonAPI.validators import is_valid_regon14

testing = {
    "REGON9": "492707333",
    "REGON14": "12345678512347"
}


@pytest.mark.first
def test_validators_Prerequisites():
    assert True


def test_is_valid_regon8_function__valid_call():
    assert is_valid_regon8("12345678")


def test_is_valid_regon8_function__invalid_calls():
    assert not is_valid_regon8("123")
    assert not is_valid_regon8(123)


def test_is_valid_regon9_function__valid_call():
    assert is_valid_regon9(testing["REGON9"])


def test_is_valid_regon9_function__invalid_calls():
    assert not is_valid_regon9("492707339")  # invalid regon9
    assert not is_valid_regon9("123")
    assert not is_valid_regon9(123)


def test_is_valid_regon13_function__valid_call():
    assert is_valid_regon13("1234567891234")


def test_is_valid_regon13_function__invalid_calls():
    assert not is_valid_regon13("123")
    assert not is_valid_regon13(123)


def test_is_valid_regon14_function__valid_call():
    assert is_valid_regon14(testing["REGON14"])


def test_is_valid_regon14_function__invalid_call():
    assert not is_valid_regon14("12345678512349")  # invalid regon14
    assert not is_valid_regon14("123")
    assert not is_valid_regon14(123)


def test_re_is_digit_string_function__valid_call():
    assert _re_is_digit_string("12345", 5)


def test_re_is_digit_string_function__invalid_calls():
    assert not _re_is_digit_string("12345", 15)
    assert not _re_is_digit_string("123abc", 3)
    assert not _re_is_digit_string("abc123", 3)
    assert not _re_is_digit_string("abc123def", 3)
    assert not _re_is_digit_string("abc", 3)
    assert not _re_is_digit_string(12345, 5)
    try:
        _re_is_digit_string("12345", "5")
        assert False
    except TypeError as e:
        assert True
    try:
        _re_is_digit_string("12345", -5)
        assert False
    except Exception as e:
        assert True



class ApiError(BaseException):
    def __str__(self):
        return repr(self.data)


class ApiAuthenticationError(ApiError):

    def __init__(self, data):
        self.data = 'Authentication failed with key: "{data}"'.format(
            data=data)


class ApiCodeTranslationError(ApiError):

    def __init__(self, data):
        self.data = 'Can\'t translate: {data}'.format(
            data=data)


class ApiUnknownReportNameError(ApiError):

    def __init__(self, data):
        self.data = 'Invailid report name: {data}'.format(
            data=data)


class ApiInvalidBIRVersionProvided(ApiError):  # TODO: test case

    def __init__(self, data, choices):
        self.data = 'Invailid BIR version: {data} not in {choices}'.format(
            data=data, choices=choices
        )



class ConvertionTypeError(BaseException):

    def __init__(self, data, what=''):
        self.data = 'Can\'t convert {what}: {data}'.format(
            data=data, what=what)

    def __str__(self):
        return repr(self.data)


class ConvertionLengthError(BaseException):

    def __init__(self, data, what=''):
        self.data = 'Invalid {what} length: {data}'.format(
            data=data, what=what)

    def __str__(self):
        return repr(self.data)


class RegonConvertionError(ConvertionTypeError):

    def __init__(self, data):
        super(RegonConvertionError, self).__init__(
            data=data, what='regon'
        )



"""
    RegonAPI main client class
"""

import sys
if sys.version_info[0] < 3:
  print("Sorry. This module doesn't support Python2")
  sys.exit(1)

import logging

from requests import Session
from zeep.transports import Transport
from zeep import Client

from .api_codes import t
from .exceptions import (
    ApiInvalidBIRVersionProvided,
    ApiAuthenticationError
)
from .operations import RegonAPIOperations
from .settings import (
    lang,
    BIR_VERSIONS,
    BIR_SETTINGS,
    REPORTS,
    DEV_ENV_WARNINGS,
    API_KEY_TEST_ENV,
    LOGGING_FORMAT
)

# logger configuration
ROOT_LOGGING_LEVEL = logging.ERROR
LOCAL_LOGGING_LEVEL = logging.WARNING
logging.basicConfig(format=LOGGING_FORMAT)
logging.root.setLevel(ROOT_LOGGING_LEVEL)
logger = logging.getLogger(__name__)
logger.setLevel(LOCAL_LOGGING_LEVEL)

# TODO: New class desc
class RegonAPI(RegonAPIOperations):
    """Regon API client

    Parameters
    ----------
    bir_version : str
        BIR version to use
    is_production : bool
        Production if True, Test if False
    service_namespace : str
        Service namespace default: {http://tempuri.org/}e3

    Attributes
    ----------
    wsdl : str
        Regon API WSDL (Web Services Description Language) URL
    service_url : str
        Service URL provided by RegonAPI Administrators
    client : zeep.client.Client
        zeep module Client object
    service : zeep.proxy.ServiceProxy
        zeep module ServiceProxy object. Used to perform SOAP requests
    key : str
        Regon API key
    sid : str
        session id
    """

    def __init__(
        self,
        bir_version="bir1.1",
        is_production=False,
        service_namespace='{http://tempuri.org/}e3'
    ):
        # Set BIR version
        self.bir_version = bir_version.lower()
        if self.bir_version not in BIR_VERSIONS:
            raise ApiInvalidBIRVersionProvided(bir_version, BIR_VERSIONS)

        # Set BIR version reports
        self.reports = REPORTS[self.bir_version]

        # Set API environment
        api_env = "PROD" if is_production else "TEST"
        if not is_production:
            logger.warning(DEV_ENV_WARNINGS["WARN_IS_NOT_PRODUCTION"][lang])


        # Set WSDL & SERVICE_URL based on BIR version
        self.wsdl = BIR_SETTINGS[bir_version][api_env]["WSDL"]
        self.service_url = BIR_SETTINGS[bir_version][api_env]["SERVICE_URL"]

        logger.debug('wsdl=%s service_url=%s' % (self.wsdl, self.service_url))

        self.client = Client(wsdl=self.wsdl)
        self.service = None
        self.key = None
        self.sid = None
        self.is_production = is_production
        self.service_namespace = service_namespace
        self._create_service()

    def authenticate(self, key, verify=True):
        """Authenticates client

        Note
        ----
        This method additionaly sets sid (session id) header in Client's
        Session object. There is no need to set it manually.

        This method additionaly sets 'key' attribute

        This method additionaly sets 'sid' attribute

        Parameters
        ----------
        key : str
            API key
        verify : bool
            If True verifies if authentication was successfull

        Returns
        -------
        str
            sid (session id) required by API for further requests.

        Raises
        ------
        ApiAuthenticationError
            If authentication was a failure
        """
        if not self.is_production:
            key = API_KEY_TEST_ENV

        logger.debug('key=%s verify=%s' % (key, verify))
        sid = self.service.Zaloguj(key)
        session = Session()
        session.headers.update({'sid': sid})
        self.client = Client(
            wsdl=self.wsdl,
            transport=Transport(session=session)
        )
        self._create_service()
        if verify:
            if not self._check_session():
                logger.error('Authentication failed')
                raise ApiAuthenticationError(key)
            else:
                logger.debug('Authenticated successfully')
        self.key = key
        self.sid = sid
        return sid

    def get_last_code(self):
        """Gets Regon API status code of last performed operation

        Returns
        -------
        tuple
            a tuple (api_response : str, translated_message: str, dict)
            translated_message is dict when settings lang='all'
        """
        response = self.service.GetValue(pNazwaParametru='KomunikatKod')
        ret = response, t('GetValue', response)
        logger.debug(ret)
        return ret

    def get_data_status(self):
        """Gets Regon API data status

        Returns
        -------
        str
            date of last database update
        """
        response = self.service.GetValue(pNazwaParametru='StanDanych')
        logger.debug(response)
        return response

    def get_service_status(self):
        """Gets Regon API status code and status message

        Returns
        -------
        tuple
            a tuple (api_response : str, translated_message: str, dict)
            translated_message is dict when settings lang='all'
        """
        response = self.service.GetValue(pNazwaParametru='StatusUslugi')
        ret = response, t('ServiceStatus', response)
        logger.debug(ret)
        return ret

    def get_operations(self):
        """Gets all operations listed in WSDL

        Returns
        -------
        list
            of names of the operations
        """
        ret = list(self.service._operations.keys())
        logger.debug("WSDL operations: %s" % ret)
        return ret

    def __str__(self):
        """Prints attributes"""
        msg = """
            WSDL        : {service_url}
            Service URL : {wsdl}
            API key     : {key}
            sid         : {sid}
            ----------------------------------
            Data status : {data_status}
        """.format(
            wsdl=self.wsdl,
            service_url=self.service_url,
            key=self.key,
            sid=self.sid,
            data_status=self.get_data_status(),
        )
        return msg

    def _create_service(self):
        """Creates service with provided URL

        Note
        ----
        This method is called after self.client if modified in
        __init__ constructor and authenticate method.
        Service is available at self.service.
        """
        logger.debug('Creating service with url %s' % self.service_url)
        self.service = self.client.create_service(
            self.service_namespace, self.service_url)


    def _check_session(self):
        """Checks Regon API for confirmation of successfull authentication

        Returns
        -------
        bool
            True if authentication successfull, else False
        """
        response = self.service.GetValue(pNazwaParametru='StatusSesji')
        return True if response == '1' else False



import setuptools
import pathlib


BASE_DIR = pathlib.Path(__file__).parent
with open(BASE_DIR/"pypi_long.md", encoding='utf-8') as fh:
    README = fh.read()


setuptools.setup(
    name="RegonAPI",
    version="1.1.1",
    author="Bartosz Nowakowski",
    author_email="rolzwy7@gmail.com",
    description="Python 3 API Client for Polish REGON database (Baza Internetowa Regon - BIR)",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/rolzwy7/RegonAPI",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Database :: Front-Ends",
        "Environment :: Console",
        ],
    install_requires=[
       "zeep",
       "beautifulsoup4",
       ]
)



# Keys are ISO 3166-2:PL abbr.
VOIVODESHIPS = {
    "DS": {
        "teryt": "02",
        "name_pl": "dolnośląskie"
    },
    "KP": {
        "teryt": "04",
        "name_pl": "kujawsko-pomorskie"
    },
    "LU": {
        "teryt": "06",
        "name_pl": "lubelskie"
    },
    "LB": {
        "teryt": "08",
        "name_pl": "lubuskie"
    },
    "LD": {
        "teryt": "10",
        "name_pl": "łódzkie"
    },
    "MA": {
        "teryt": "12",
        "name_pl": "małopolskie"
    },
    "MZ": {
        "teryt": "14",
        "name_pl": "mazowieckie"
    },
    "OP": {
        "teryt": "16",
        "name_pl": "opolskie"
    },
    "PK": {
        "teryt": "18",
        "name_pl": "podkarpackie"
    },
    "PD": {
        "teryt": "20",
        "name_pl": "podlaskie"
    },
    "PM": {
        "teryt": "22",
        "name_pl": "pomorskie"
    },
    "SL": {
        "teryt": "24",
        "name_pl": "śląskie"
    },
    "SK": {
        "teryt": "26",
        "name_pl": "świętokrzyskie"
    },
    "WN": {
        "teryt": "28",
        "name_pl": "warmińsko-mazurskie"
    },
    "WP": {
        "teryt": "30",
        "name_pl": "wielkopolskie"
    },
    "ZP": {
        "teryt": "32",
        "name_pl": "zachodniopomorskie"
    },
}



import os
import logging

from .consts import *

# Language - settings
lang = 'eng'

# Logger - settings
available_languages = ["pl", "eng"]

LOGGING_FORMAT = "\r%(asctime)s, %(levelname)-8s [`%(funcName)s` %(filename)s:%(lineno)d] %(message)s"


from unittest import mock
import pytest

from RegonAPI import RegonAPI
from RegonAPI.exceptions import ApiCodeTranslationError
from RegonAPI.exceptions import ApiInvalidBIRVersionProvided
from RegonAPI.exceptions import ApiAuthenticationError
from RegonAPI.settings import BIR_VERSIONS

testing = {
    "SID": "3xg5b1m7nuoeye55h667",
    "KEY": "abcde12345abcde12345"
}


@pytest.fixture
@mock.patch('RegonAPI.regon_api.Client', autospec=True)
def api_mock(mockClient):
    """ Creates api object with mocked client """
    def fin():
        pass
    api = RegonAPI()
    return api


@pytest.mark.first
def test_regon_api_Prerequisites():
    assert True


def test_get_last_code_CorrectResponse_ReturnTypeCorrect(api_mock):
    api_mock.service.GetValue.return_value = '4'
    code, msg = api_mock.get_last_code()
    api_mock.service.GetValue.assert_called_once()
    api_mock.service.GetValue.assert_called_with(
        pNazwaParametru='KomunikatKod')
    assert code == '4'
    assert isinstance(msg, str) or isinstance(msg, dict)


def test_get_last_IncorrectResponse_ExceptionRaised(api_mock):
    api_mock.service.GetValue.reset_mock()
    try:
        api_mock.service.GetValue.return_value = 'testing'
        code, msg = api_mock.get_last_code()
        assert False
    except ApiCodeTranslationError as e:
        assert True
        api_mock.service.GetValue.assert_called_once()
        api_mock.service.GetValue.assert_called_with(
            pNazwaParametru='KomunikatKod')
    except Exception:
        assert False


def test_get_data_status_CorrectResponse_ReturnTypeCorrect(api_mock):
    api_mock.service.GetValue.return_value = '2000-10-10'
    response = api_mock.get_data_status()
    api_mock.service.GetValue.assert_called_once()
    api_mock.service.GetValue.assert_called_with(
        pNazwaParametru='StanDanych')
    assert isinstance(response, str)
    assert response == '2000-10-10'


def test_get_service_status_CorrectResponse_ReturnTypeCorrect(api_mock):
    api_mock.service.GetValue.return_value = '1'
    code, msg = api_mock.get_service_status()
    api_mock.service.GetValue.assert_called_once()
    api_mock.service.GetValue.assert_called_with(
        pNazwaParametru='StatusUslugi')
    assert code == '1'
    assert isinstance(msg, str) or isinstance(msg, dict)


def test_get_service_status_IncorrectResponse_ExceptionRaised(api_mock):
    api_mock.service.GetValue.reset_mock()
    try:
        api_mock.service.GetValue.return_value = 'testing'
        code, msg = api_mock.get_service_status()
        assert False
    except ApiCodeTranslationError as e:
        assert True
        api_mock.service.GetValue.assert_called_once()
        api_mock.service.GetValue.assert_called_with(
            pNazwaParametru='StatusUslugi')


def test_get_operations_OwnOperations_NoProblemsExpected(api_mock):
    api_mock.service._operations = {"a": 1, "b": 2, "c": 3}
    operations = api_mock.get_operations()
    assert operations == ["a", "b", "c"]

    api_mock.service._operations = {}
    operations = api_mock.get_operations()
    assert operations == []


def test_get_operations_NoOperatoions_ExceptionRaised(api_mock):
    api_mock.service._operations = None
    try:
        operations = api_mock.get_operations()
        assert False
    except AttributeError:
        assert True
    except Exception:
        assert False


@mock.patch('RegonAPI.regon_api.Client')
@mock.patch('RegonAPI.regon_api.Session')
@mock.patch('RegonAPI.regon_api.RegonAPI._check_session')
@mock.patch('RegonAPI.regon_api.RegonAPI._create_service')
def test_regon_api_auth_success(
    mocked_create_service, mocked_check_session,
    mocked_Session, mocked_Client, api_mock
):
    api_mock.service.Zaloguj.return_value = testing["SID"]
    mocked_check_session.return_value = True

    try:
        returned = api_mock.authenticate(testing["KEY"], verify=True)
        assert True
    except ApiAuthenticationError as e:
        assert False

    api_mock.service.Zaloguj.assert_called_once()
    api_mock.service.Zaloguj.assert_called_with(testing["KEY"])

    mocked_Session.assert_called_once()
    mocked_Client.assert_called_once()
    mocked_create_service.assert_called_once()

    assert api_mock.key == testing["KEY"]
    assert api_mock.sid == testing["SID"]
    assert returned == testing["SID"]


@mock.patch('RegonAPI.regon_api.Client')
@mock.patch('RegonAPI.regon_api.Session')
@mock.patch('RegonAPI.regon_api.RegonAPI._check_session')
@mock.patch('RegonAPI.regon_api.RegonAPI._create_service')
def test_regon_api_auth_failure(
    mocked_create_service, mocked_check_session, mocked_Client,
    mocked_Session, api_mock
):
    api_mock.service.Zaloguj.return_value = 'testing'
    mocked_check_session.return_value = False
    try:
        api_mock.authenticate(testing["KEY"], verify=True)
        assert False
    except ApiAuthenticationError as e:
        assert True


def test_CorrectInit_NoProblemsExpected(api_mock):
    api_mock.service.GetValue.return_value = '2000-10-10'
    try:
        str_ = str(api_mock)
        assert True
    except Exception as e:
        assert False
    assert isinstance(str_, str)
    assert str_ != ''
    api_mock.service.GetValue.assert_called_once()
    api_mock.service.GetValue.assert_called_with(
        pNazwaParametru='StanDanych')


def test_CorrectParamsProvided_NoProblemsExpected(api_mock):
    # attr service_namespace
    assert isinstance(api_mock.service_namespace, str)
    # attr bir_version
    assert isinstance(api_mock.bir_version, str)
    assert(api_mock.bir_version in BIR_VERSIONS)
    # attr wsdl
    assert isinstance(api_mock.wsdl, str)
    # attr service_url
    assert isinstance(api_mock.service_url, str)

    assert api_mock.service is not None
    assert api_mock.key is None
    assert api_mock.sid is None

    api_mock.client.create_service.assert_called_once()
    api_mock.client.create_service.assert_called_once_with(
        api_mock.service_namespace,
        api_mock.service_url
    )


def test_IncorrectBirVersionProvided_ExceptionRaised():
    try:
        RegonAPI(bir_version="testing")
        assert False
    except ApiInvalidBIRVersionProvided as e:
        assert True
    except Exception as e:
        assert False
